# Generated by Django 4.0.5 on 2022-06-08 17:59

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0015_factions"),
        ("memberaudit", "0008_add_compliance_groups"),
    ]

    operations = [
        migrations.AddField(
            model_name="character",
            name="eve_character",
            field=models.OneToOneField(
                blank=True,
                default=None,
                help_text="ownership of this character on Auth",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="memberaudit_character",
                to="eveonline.evecharacter",
            ),
        ),
    ]
