{% extends 'admin/base_site.html' %}

{% block content %}

<form action="" method="post">
    {% csrf_token %}
    {{ form.as_p }}
    <br>
    <br>
    <input type="submit" value="Submit">
    <a href="{% url 'admin:memberaudit_skillset_changelist' %}">
        <input type="button" name="Cancel" value="Cancel">
    </a>
</form>

{% endblock %}
