# Generated by Django 4.0.5 on 2022-06-08 18:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("memberaudit", "0010_update_evecharacter_field"),
    ]

    operations = [
        migrations.AlterField(
            model_name="character",
            name="character_ownership",
            field=models.AutoField(primary_key=True, serialize=False),
        ),
    ]
